@@ -127,7 +127,7 @@ post_request()
 
     dsgw_emits( "\n<FONT SIZE=+1>\n" );
 
-    rdns = ldap_explode_dn( dn, 1 );
+    rdns = dsgw_ldap_explode_dn( dn, 1 );
     if ( rdns == NULL || rdns[ 0 ] == NULL ) {
 	entry_name = dn;
     } else {
@@ -136,7 +136,7 @@ post_request()
     new_name = success_msg = "";
     dsgw_emitf( "%s <B>%s</B>...\n</FONT>\n\n", changedesc, entry_name );
     if ( rdns != NULL ) {
-	ldap_value_free( rdns );
+	dsgw_charray_free( rdns );
     }
 
     if ( verbose ) {
@@ -220,7 +220,7 @@ post_request()
 	if ( rc == LDAP_SUCCESS ) {
 
 	    /* construct the new DN so we can insert correct "edit" link */
-	    if (( oldrdns = ldap_explode_dn( dn, 0 )) == NULL ) {
+	    if (( oldrdns = dsgw_ldap_explode_dn( dn, 0 )) == NULL ) {
 		dsgw_error( DSGW_ERR_NOMEMORY, NULL, DSGW_ERROPT_EXIT,
 			0, NULL );
 	    }
@@ -236,17 +236,17 @@ post_request()
 		strcat( dn, "," );
 		strcat( dn, oldrdns[ i ] );
 	    }
-	    ldap_value_free( oldrdns );
+	    dsgw_charray_free( oldrdns );
 	    free( encodeddn );
 	    encodeddn = dsgw_strdup_escaped( dn );
 
 	    success_msg = XP_GetClientStr(DBT_renamedBSBToBSB_);
-	    if (( rdns = ldap_explode_rdn( newrdn, 1 )) == NULL
+	    if (( rdns = dsgw_ldap_explode_rdn( newrdn, 1 )) == NULL
 		    || rdns[ 0 ] == NULL ) {
 		new_name = newrdn;
 	    } else {
 		new_name = dsgw_ch_strdup (rdns[ 0 ]);
-		ldap_value_free( rdns );
+		dsgw_charray_free( rdns );
 	    }
 	}
 	break;
@@ -362,7 +362,7 @@ post_request()
     if ( !quiet ) {
 	dsgw_html_end();
     }
-    ldap_unbind( ld );
+    ldap_unbind_ext( ld, NULL, NULL );
     if (old_dn != dn) free ( old_dn );
     free( dn );
 }
@@ -607,10 +607,10 @@ entry_modify_or_add( LDAP *ld, char *dn, int add, int *pwdchangedp )
 
 	if( lderr == LDAP_SUCCESS ) {
 		if(( lderr = ldap_result( ld, msgid, 1, (struct timeval *)NULL, &res )) == -1 ) {
-			lderr = ldap_get_lderrno( ld, NULL, &errmsg );
+			lderr = dsgw_ldap_get_lderrno( ld, NULL, &errmsg );
 			modify_error( lderr, errmsg );
 		} else {
-			lderr = ldap_result2error( ld, res, 1 ); 
+			lderr = ldap_parse_result( ld, res, NULL, NULL, NULL, NULL, NULL, 1 ); 
 			if ( lderr == LDAP_SUCCESS ) {
 				if ( !quiet ) {
 				if ( add ) {
@@ -620,7 +620,7 @@ entry_modify_or_add( LDAP *ld, char *dn, int add, int *pwdchangedp )
 				}
 				}
 			} else {
-				(void)ldap_get_lderrno( ld, NULL, &errmsg );
+				(void)dsgw_ldap_get_lderrno( ld, NULL, &errmsg );
 				modify_error( lderr, errmsg );
 
 				/* Do some checks for password policy infractions. */
@@ -633,7 +633,7 @@ entry_modify_or_add( LDAP *ld, char *dn, int add, int *pwdchangedp )
 			}
 		}
 	} else {
-		(void)ldap_get_lderrno( ld, NULL, &errmsg );
+		(void)dsgw_ldap_get_lderrno( ld, NULL, &errmsg );
 		modify_error( lderr, errmsg );
 	}
 
@@ -652,12 +652,12 @@ entry_delete( LDAP *ld, char *dn )
     char	*errmsg = NULL;
 
     dsgw_emitf( "<FONT SIZE=+1>\n" );
-    if (( lderr = ldap_delete_s( ld, dn )) == LDAP_SUCCESS ) {
+    if (( lderr = ldap_delete_ext_s( ld, dn, NULL, NULL )) == LDAP_SUCCESS ) {
 	if ( !quiet ) {
 	    dsgw_emitf( XP_GetClientStr(DBT_PSuccessfullyDeletedEntryN_) );
 	}
     } else {
-	(void)ldap_get_lderrno( ld, NULL, &errmsg );
+	(void)dsgw_ldap_get_lderrno( ld, NULL, &errmsg );
 	modify_error( lderr, errmsg );
     }
 
@@ -678,13 +678,13 @@ entry_modrdn( LDAP *ld, char *dn, char *newrdn, int deleteoldrdn )
     }
 
     dsgw_emitf( "<FONT SIZE=+1>\n" );
-    if (( lderr = ldap_modrdn2_s( ld, dn, newrdn, deleteoldrdn ))
+    if (( lderr = ldap_rename_s( ld, dn, newrdn, NULL, deleteoldrdn, NULL, NULL ))
 	    == LDAP_SUCCESS ) {
 	if ( !quiet ) {
 	    dsgw_emitf( XP_GetClientStr(DBT_PSuccessfullyRenamedEntryN_) );
 	}
     } else {
-	(void)ldap_get_lderrno( ld, NULL, &errmsg );
+	(void)dsgw_ldap_get_lderrno( ld, NULL, &errmsg );
 	modify_error( lderr, errmsg );
     }
 
@@ -994,12 +994,13 @@ value_is_unique( LDAP *ld, char *dn, char *attr, char *value )
     attrs[ 0 ] = attr;
     attrs[ 1 ] = NULL;
 
-    rc = ldap_search_s( ld, gc->gc_ldapsearchbase, LDAP_SCOPE_SUBTREE,
-	    buf, attrs, 1, &res );
+    rc = ldap_search_ext_s( ld, gc->gc_ldapsearchbase, LDAP_SCOPE_SUBTREE, 
+                            buf, attrs, 1, NULL, NULL, NULL /* no timelimit */,
+                            -1 /* no sizelimit */, &res );
     free( buf );
 
     if ( rc != LDAP_SUCCESS || res == NULL ) {
-	(void)ldap_get_lderrno( ld, NULL, &errmsg );
+	(void)dsgw_ldap_get_lderrno( ld, NULL, &errmsg );
 	modify_error( rc, errmsg );
 	return( rc );
     }
@@ -1011,7 +1012,7 @@ value_is_unique( LDAP *ld, char *dn, char *attr, char *value )
     } else {	/* found one entry: see if it is the entry we are modifying */
 	if (( e = ldap_first_entry( ld, res )) == NULL ||
 	    ( tmpdn = ldap_get_dn( ld, e )) == NULL ) {
-	    rc = ldap_get_lderrno( ld, NULL, NULL );
+	    rc = dsgw_ldap_get_lderrno( ld, NULL, NULL );
 	} else if ( dsgw_dn_cmp( dn, tmpdn ) != 0 ) {
 	    rc = LDAP_SUCCESS;	/* same entry */
 	} else {
