@@ -108,7 +108,7 @@ static unsigned long attroptvals[] = {
 	DSGW_ATTROPT_ENTITIES,
 };
 
-
+#define DSGW_ATTRARG_DN		"dn"
 #define DSGW_ATTRARG_ATTR	"attr"
 #define DSGW_ATTRARG_SYNTAX	"syntax"
 #define DSGW_ATTRARG_HTMLTYPE	"type"
@@ -153,8 +153,8 @@ struct dsgw_attrdispinfo {
     char		*adi_attr;
     int			adi_argc;
     char		**adi_argv;
-    char		**adi_vals;
-    char		*adi_rdn; /* a copy of adi_vals[i] (possibly NULL) */
+    struct berval	**adi_vals;
+    struct berval	*adi_rdn; /* a copy of adi_vals[i] (possibly NULL) */
     int			adi_htmltype;
     unsigned long	adi_opts;
 };
@@ -179,14 +179,15 @@ static void output_prelude( dsgwtmplinfo *tip );
 static void output_nonentry_line( dsgwtmplinfo *tip, char *line );
 static struct attr_handler *syntax2attrhandler( char *syntax );
 static int numfields( int argc, char **argv, int valcount );
-static void element_sizes( int argc, char **argv, char **vals, int valcount,
-	int *rowsp, int *colsp );
+static void element_sizes( int argc, char **argv, struct berval **vals, 
+	int valcount, int *rowsp, int *colsp );
 #define DSGW_TEXTOPT_FOCUSHANDLERS	0x0001
 #define DSGW_TEXTOPT_CHANGEHANDLERS	0x0002
-static void output_text_elements( int argc, char **argv, char *attr,
-	char **vals, const char* rdn, char *prefix, int htmltype, unsigned long opts );
-static void output_textarea( int argc, char **argv, char *attr,
-	char **vals, int valcount, char *prefix, unsigned long opts );
+static void output_text_elements( int argc, char **argv, char *attr, 
+	struct berval **vals, const char* rdn, char *prefix, int htmltype, 
+	unsigned long opts );
+static void output_textarea( int argc, char **argv, char *attr, 
+	struct berval **vals, int valcount, char *prefix, unsigned long opts );
 static void emit_value( char *val, int quote_html_specials );
 static void output_text_checkbox_or_radio( struct dsgw_attrdispinfo *adip,
 	char *prefix, int htmltype );
@@ -207,16 +208,15 @@ static void do_passwordfield( unsigned long dispopts, int argc, char **argv,
 static void do_helpbutton( unsigned long dispopts, int argc, char **argv );
 static void do_closebutton( unsigned long dispopts, int argc, char **argv );
 static void do_viewswitcher( char *template, char *dn, int argc, char **argv );
-static int did_output_as_special( int argc, char **argv, char *label,
-	char *val );
+static int did_output_as_special( int argc, char **argv, struct berval *label,
+	struct berval *val );
 static char *time2text( char *ldtimestr, int dateonly );
 static long gtime( struct tm *tm );
 static int looks_like_dn( char *s );
 static void do_std_completion_js( char *template, int argc, char **argv );
 static int condition_is_true( int argc, char **argv, void *arg );
-static char ** dsgw_get_values( LDAP *ld, LDAPMessage *entry, 
-				const char *target, int binary_value );
-static void dsgw_value_free( void **ldvals, int binary ) ;
+static struct berval ** dsgw_get_values( LDAP *ld, LDAPMessage *entry, 
+				const char *target );
 static char *dsgw_time(time_t secs_since_1970);
 
 /* attribute syntax handler routines */
@@ -693,13 +693,13 @@ dsgw_display_done( dsgwtmplinfo *tip, char *dn )
     dsgw_savelines_free( tip->dsti_entrylines );
     fclose( tip->dsti_fp );
     if ( tip->dsti_attrs != NULL ) {
-	ldap_value_free( tip->dsti_attrs );
+	dsgw_charray_free( tip->dsti_attrs );
     }
     if ( tip->dsti_attrflags != NULL ) {
 	free( tip->dsti_attrflags );
     }
     if ( tip->dsti_rdncomps != NULL ) {
-	ldap_value_free( tip->dsti_rdncomps );
+	dsgw_charray_free( tip->dsti_rdncomps );
     }
     free( tip );
 }
@@ -897,30 +897,30 @@ output_nonentry_line( dsgwtmplinfo *tip, char *line )
     }
 }
 
-static char*
-find_RDN (char* DN, char* attr, char** vals)
-     /* Return a copy of the vals[i] that is
+static struct berval *
+find_RDN (char* DN, char* attr, struct berval** vals)
+     /* Return a pointer to the vals[i] that is
 	part of the RDN of the given DN.
      */
 {
     if (DN && *DN && vals && *vals) {
-	auto char** RDNs = ldap_explode_dn (DN, 0);
-	auto char** AVAs = ldap_explode_rdn (RDNs[0], 0);
-	ldap_value_free (RDNs);
+	auto char** RDNs = dsgw_ldap_explode_dn (DN, 0);
+	auto char** AVAs = dsgw_ldap_explode_rdn (RDNs[0], 0);
+	dsgw_charray_free (RDNs);
 	if (AVAs) {
-	    auto char** val = NULL;
+	    struct berval **val = NULL;
 	    auto char** AVA;
 	    for (AVA = AVAs; *AVA; ++AVA) {
 		auto char* RDN = strchr (*AVA, '=');
 		if (RDN) {
 		    *RDN++ = '\0';
 		    if (!strcasecmp (*AVA, attr)) {
-			for (val = vals; *val; ++val) {
-			    if (!strcmp (RDN, *val)) {
+			for (val = vals; *val && (*val)->bv_val; ++val) {
+			    if (!strcmp (RDN, (*val)->bv_val)) {
 				break;
 			    }
 			}
-			if (*val) break;
+			if (*val && (*val)->bv_val) break;
 			/* bug: what if there are other AVAs
 			   that also match attr and one of vals?
 			   Even if this algorithm could find them,
@@ -931,41 +931,30 @@ find_RDN (char* DN, char* attr, char** vals)
 		    }
 		}
 	    }
-	    ldap_value_free (AVAs);
+	    dsgw_charray_free (AVAs);
 	    if (val) return *val;
 	}
     }
     return NULL;
 }
 
-/*static int
- *is_aim_online(dsgwtmplinfo *tip) 
- *{
- * char **ldvals = (char **) dsgw_get_values(tip->dsti_ld, tip->dsti_entry, DSGW_ATTRTYPE_AIMSTATUSTEXT, 0);
- *
- * if (ldvals == NULL || *ldvals == NULL || strcmp(*ldvals, "") == 0 ) {
- *   return(0);
- * }
- * return(1);
- *
- *}
- */
 static void
 do_orgchartlink( dsgwtmplinfo *tip, char *dn, unsigned long dispopts,
 	int argc, char **argv )
 {
-  char **ldvals = (char **) dsgw_get_values(tip->dsti_ld, tip->dsti_entry, gc->gc_orgchartsearchattr, 0);
+  struct berval **ldvals = dsgw_get_values(tip->dsti_ld, tip->dsti_entry, gc->gc_orgchartsearchattr);
   char *escaped_value;
 
-  if (gc->gc_orgcharturl == NULL || ldvals == NULL || *ldvals == NULL || strcmp(*ldvals,"") == 0) {
+  if (gc->gc_orgcharturl == NULL || ldvals == NULL || *ldvals == NULL ||
+      (*ldvals)->bv_val == NULL || strcmp((*ldvals)->bv_val, "") == 0) {
     dsgw_emits("\"javascript:void(0)\"");
     return;
   }
   dsgw_emits("\"");
   dsgw_emits(gc->gc_orgcharturl);
-  escaped_value = dsgw_ch_malloc( 3 * strlen( ldvals[0] ) + 1 );
+  escaped_value = dsgw_ch_malloc( 3 * strlen( ldvals[0]->bv_val ) + 1 );
   *escaped_value = '\0';
-  dsgw_strcat_escaped( escaped_value, ldvals[0]);
+  dsgw_strcat_escaped( escaped_value, ldvals[0]->bv_val);
   dsgw_emits(escaped_value);
   dsgw_emits("\"\n");
 
@@ -976,14 +965,12 @@ static void
 do_attribute( dsgwtmplinfo *tip, char *dn, unsigned long dispopts,
 	int argc, char **argv )
 {
-    char			*attr, *syntax, *defval, *tmpvals[ 2 ], *s;
-    char			**ldvals, **vals;
+    char			*attr, *syntax, *defval, *s;
+    struct berval		**ldvals, **vals, tmpval, *tmpvals[ 2 ];
     unsigned long		options;
     int				i, len, attrindex, htmltype;
     struct dsgw_attrdispinfo	adi;
     int                         editable = 0;
-    int                         tagged_attrs = 0;
-    int                         binary_value = 0;
     
     if (( attr = get_arg_by_name( DSGW_ATTRARG_ATTR, argc, argv )) == NULL ) {
 	dsgw_emitf( XP_GetClientStr(DBT_missingS_), DSGW_ATTRARG_ATTR );
@@ -1076,42 +1063,27 @@ do_attribute( dsgwtmplinfo *tip, char *dn, unsigned long dispopts,
     ldvals = vals = NULL;
 
     if ( strcasecmp( attr, "dn" ) == 0 ) {	/* dn pseudo-attribute */
-	tmpvals[ 0 ] = dn;
+	tmpval.bv_val = dn;
+	tmpval.bv_len = strlen(dn);
+	tmpvals[ 0 ] = &tmpval;
 	tmpvals[ 1 ] = NULL;
 	vals = tmpvals;
 	options &= ~DSGW_ATTROPT_EDITABLE;	/* always read-only */
 	options &= ~DSGW_ATTROPT_ADDING;	/* always read-only */
 	options |= DSGW_ATTROPT_READONLY;
-	} else if( strcasecmp( syntax, "binvalue" ) == 0) {
-
-	    binary_value = 1;
-	    /* Only display tagged stuff on searches */
-	    if (editable){
-		ldvals = (char **) ldap_get_values_len(tip->dsti_ld, tip->dsti_entry, attr);
-		tagged_attrs = 0;
-	    } else {
-		ldvals = (char **) dsgw_get_values(tip->dsti_ld, tip->dsti_entry, attr, 1 /*binary value*/);
-		tagged_attrs = 1;
-	    }
-
-	    if (ldvals != NULL) {
-		vals = ldvals;
-	    }
-	} else if ( tip->dsti_entry != NULL) {
-
-	    /* Only display tagged stuff on searches */
-	    if ( editable){
-		ldvals = (char **) ldap_get_values( tip->dsti_ld, tip->dsti_entry, attr);
-		tagged_attrs = 0;
-	    } else {
-		ldvals = (char **) dsgw_get_values( tip->dsti_ld, tip->dsti_entry, attr, 0 );
-		tagged_attrs = 1;
-	    }
-	    if (ldvals != NULL) {
-		vals = ldvals;
-	    }
+    } else if( (strcasecmp( syntax, "binvalue" ) == 0) || (tip->dsti_entry != NULL) )  {
+	/* Only display tagged stuff on searches */
+	if (editable){
+	    ldvals = ldap_get_values_len(tip->dsti_ld, tip->dsti_entry, attr);
+	} else {
+	    ldvals = dsgw_get_values(tip->dsti_ld, tip->dsti_entry, attr);
 	}
 
+	if (ldvals != NULL) {
+	    vals = ldvals;
+	}
+    }
+
     if (vals == NULL && (options & DSGW_ATTROPT_QUOTED ) != 0 ) {
       dsgw_emits( "\"\"" );
       return;
@@ -1129,7 +1101,9 @@ do_attribute( dsgwtmplinfo *tip, char *dn, unsigned long dispopts,
 	    if (( s = strchr( tip->dsti_rdncomps[ i ], '=' )) != NULL &&
 		    s - tip->dsti_rdncomps[ i ] == len &&
 		    strncasecmp( attr, tip->dsti_rdncomps[ i ], len ) == 0 ) {
-		tmpvals[ 0 ] = ++s;
+		tmpval.bv_val = ++s;
+		tmpval.bv_len = strlen(tmpval.bv_val);
+		tmpvals[ 0 ] = &tmpval;
 		tmpvals[ 1 ] = NULL;
 		vals = tmpvals;
 		break;
@@ -1139,7 +1113,9 @@ do_attribute( dsgwtmplinfo *tip, char *dn, unsigned long dispopts,
 
     if ( vals == NULL && ( defval = get_arg_by_name( DSGW_ATTRARG_DEFAULT,
 	    argc, argv )) != NULL ) {
-	tmpvals[ 0 ] = defval;
+	tmpval.bv_val = defval;
+	tmpval.bv_len = strlen(tmpval.bv_val);
+	tmpvals[ 0 ] = &tmpval;
 	tmpvals[ 1 ] = NULL;
 	vals = tmpvals;
     }
@@ -1154,8 +1130,8 @@ do_attribute( dsgwtmplinfo *tip, char *dn, unsigned long dispopts,
 	} else {
 	    if ( vals != NULL && vals[1] != NULL
 			&& ( options & DSGW_ATTROPT_SORT ) != 0 ) {
-		ldap_sort_values( tip->dsti_ld, vals,
-				 dsgw_valcmp (adi.adi_handlerp->ath_compare));
+		dsgw_sort_values( vals,
+				  dsgw_valcmp (adi.adi_handlerp->ath_compare));
 	    }
 	    adi.adi_attr = attr;
 	    adi.adi_argc = argc;
@@ -1193,21 +1169,9 @@ do_attribute( dsgwtmplinfo *tip, char *dn, unsigned long dispopts,
 	}
     }
 
-    if ( ldvals != NULL ) {
-	if (tagged_attrs) {
-	    dsgw_value_free( (void **) ldvals, binary_value );
-	} else {
-	    if (binary_value) {
-		ldap_value_free_len( (struct berval **) ldvals );
-	    } else {
-		ldap_value_free( ldvals );
-	    }
-	}
-    }
+    ldap_value_free_len(ldvals);
 }
 
-
-
 static void
 append_to_array( char ***ap, int *countp, char *s )
 {
@@ -1309,7 +1273,7 @@ numfields( int argc, char **argv, int valcount )
  * the number of values, and the length of longest value.
  */
 static void
-element_sizes( int argc, char **argv, char **vals, int valcount,
+element_sizes( int argc, char **argv, struct berval **vals, int valcount,
 	int *rowsp, int *colsp )
 {
     int		i, len, maxlen;
@@ -1340,7 +1304,7 @@ element_sizes( int argc, char **argv, char **vals, int valcount,
 	    /* determine ( length of longest value ) + 1  */
 	    maxlen = 0;
 	    for ( i = 0; i < valcount; ++i ) {
-		if (( len = strlen( vals[ i ] )) > maxlen ) {
+		if (( len = strlen( vals[ i ]->bv_val )) > maxlen ) {
 		    maxlen = len;
 		}
 	    }
@@ -1383,7 +1347,7 @@ element_sizes( int argc, char **argv, char **vals, int valcount,
 
 
 static void
-output_text_elements( int argc, char **argv, char *attr, char **vals,
+output_text_elements( int argc, char **argv, char *attr, struct berval **vals,
 	const char* rdn, char *prefix, int htmltype, unsigned long opts )
 {
     int		i, valcount, fields, cols;
@@ -1391,16 +1355,16 @@ output_text_elements( int argc, char **argv, char *attr, char **vals,
     if ( vals == NULL ) {
 	valcount = 0;
     } else {
-	for ( valcount = 0; vals[ valcount ] != NULL; ++valcount ) {
+	for ( valcount = 0; vals[ valcount ] && vals[ valcount ]->bv_val; ++valcount ) {
 		/* just count vals  */
 	}
-	}
+    }
 
     fields = numfields( argc, argv, valcount );
     element_sizes( argc, argv, vals, valcount, NULL, &cols );
 
     for ( i = 0; i < fields; ++i ) {
-	auto const int is_rdn = (i < valcount && vals[ i ] == rdn);
+	auto const int is_rdn = (i < valcount && vals[ i ]->bv_val == rdn);
 
 	dsgw_emitf( "<INPUT TYPE=\"%s\"", attrhtmltypes[ htmltype ] );
 
@@ -1410,7 +1374,7 @@ output_text_elements( int argc, char **argv, char *attr, char **vals,
 	}
 	
 	if ( i < valcount ) {
-	    dsgw_emitf( " VALUE=\"%s\"", vals[ i ] );
+	    dsgw_emitf( " VALUE=\"%s\"", vals[ i ]->bv_val );
 	}
 	
 	if (( opts & DSGW_TEXTOPT_CHANGEHANDLERS ) != 0 ) {
@@ -1429,7 +1393,7 @@ output_text_elements( int argc, char **argv, char *attr, char **vals,
 
 
 static void
-output_textarea( int argc, char **argv, char *attr, char **vals,
+output_textarea( int argc, char **argv, char *attr, struct berval **vals,
 	int valcount, char *prefix, unsigned long opts )
 {
     int		i, rows, cols;
@@ -1459,7 +1423,7 @@ output_textarea( int argc, char **argv, char *attr, char **vals,
     dsgw_emits( ">\n" );
 
     for ( i = 0; i < valcount; ++i ) {
-	dsgw_emits( vals[ i ] );
+	dsgw_emits( vals[ i ]->bv_val );
 	dsgw_emits( "\n" );
     }
 
@@ -1472,7 +1436,12 @@ output_text_checkbox_or_radio( struct dsgw_attrdispinfo *adip, char *prefix,
 	int htmltype )
 {
     int		i, checked;
+#if 0
     char	*value;
+#else
+    struct berval value;
+    struct berval *valuep = &value;
+#endif
 
     /*
      * for checkboxes or radio buttons that are associated with string values,
@@ -1480,12 +1449,14 @@ output_text_checkbox_or_radio( struct dsgw_attrdispinfo *adip, char *prefix,
      * present.
      */
     checked = 0;
-    if (( value = get_arg_by_name( DSGW_ATTRARGS_VALUE, adip->adi_argc,
+    if (( value.bv_val = get_arg_by_name( DSGW_ATTRARGS_VALUE, adip->adi_argc,
 	    adip->adi_argv )) == NULL ) {
-	value = "TRUE";	/* assume LDAP Boolean value */
+	value.bv_val = "TRUE";	/* assume LDAP Boolean value */
     }
+    /* value.bv_val never be NULL */
+    value.bv_len = strlen(value.bv_val);
     if ( adip->adi_vals == NULL ) {
-	if ( *value == '\0' ) {
+	if ( *(value.bv_val) == '\0' ) {
 	    /*
 	     * There are no existing values in the entry and this checkbox or
 	     * radio button has a zero-length value associated with it.  We
@@ -1496,9 +1467,10 @@ output_text_checkbox_or_radio( struct dsgw_attrdispinfo *adip, char *prefix,
 	}
 
     } else {
-	for ( i = 0; adip->adi_vals[ i ] != NULL; ++i ) {
-	    if ( dsgw_valcmp(adip->adi_handlerp->ath_compare)( (const char **)&value,
-		    (const char **)&(adip->adi_vals[ i ]) ) == 0 ) {
+	for ( i = 0; adip->adi_vals[ i ] && adip->adi_vals[ i ]->bv_val; ++i ) {
+	    if ( dsgw_valcmp(adip->adi_handlerp->ath_compare)( 
+		       (const struct berval **)&valuep,
+		       (const struct berval **)&(adip->adi_vals[ i ]) ) == 0 ) {
 		checked = 1;
 		break;
 	    }
@@ -1507,7 +1479,7 @@ output_text_checkbox_or_radio( struct dsgw_attrdispinfo *adip, char *prefix,
     dsgw_emitf( "<INPUT TYPE=\"%s\" NAME=\"%s%s\" "
 	    "VALUE=\"%s\"%s onClick=\"aChg('%s')\">\n",
 	    ( htmltype == DSGW_ATTRHTML_RADIO ) ? "radio" : "checkbox",
-	    prefix, adip->adi_attr, value, checked ? " CHECKED" : "",
+	    prefix, adip->adi_attr, value.bv_val, checked ? " CHECKED" : "",
 	    adip->adi_attr );
 }
 
@@ -1537,13 +1509,13 @@ emit_value( char *val, int quote_html_specials )
 static void
 binvalue_display( struct dsgw_attrdispinfo *adip )
 {
-    int		i;
-    struct berval **list_of_binvals;
+	int		i;
+	struct berval **list_of_binvals;
 	char *checked = " CHECKED";
 	char *selected = " SELECTED";
 	int		iValue;
 
-	list_of_binvals = (struct berval **)adip->adi_vals;
+	list_of_binvals = adip->adi_vals;
 
 	for ( i = 0; list_of_binvals[ i ] != NULL; ++i ) 
 	{
@@ -1662,23 +1634,23 @@ ntdomain_display( struct dsgw_attrdispinfo *adip )
 
     /* Write values with a break (<BR>) separating them, 
 	removing all after ":" */
-    for ( i = 0; adip->adi_vals[ i ] != NULL; ++i ) {
+    for ( i = 0; adip->adi_vals[ i ] && adip->adi_vals[ i ]->bv_val; ++i ) {
 	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv, 
 		adip->adi_vals[ i ], adip->adi_vals[ i ] )) {
-            char *pch = strchr( adip->adi_vals[ i ], DSGW_NTDOMAINID_SEP );
+            char *pch = strchr( adip->adi_vals[ i ]->bv_val, DSGW_NTDOMAINID_SEP );
             if( pch )
                *pch = (char )NULL;
 	    if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		dsgw_emits( "\"" );
 	    }
 	    
-	    fputs( adip->adi_vals[ i ], stdout );
+	    fputs( adip->adi_vals[ i ]->bv_val, stdout );
 	    if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		dsgw_emits( "\"" );
 	    }
 	}
 
-	if ( adip->adi_vals[ i + 1 ] != NULL ) {
+	if ( adip->adi_vals[ i + 1 ] && adip->adi_vals[ i + 1 ]->bv_val ) {
 	    fputs( "<BR>\n", stdout );
 	}
     }
@@ -1702,7 +1674,7 @@ str_display( struct dsgw_attrdispinfo *adip )
     }
 
     /* just write values with a break (<BR>) separating them */
-    for ( i = 0; adip->adi_vals[ i ] != NULL; ++i ) {
+    for ( i = 0; adip->adi_vals[ i ] && adip->adi_vals[ i ]->bv_val; ++i ) {
 
 	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv,
 		adip->adi_vals[ i ], adip->adi_vals[ i ] ) &&
@@ -1710,7 +1682,7 @@ str_display( struct dsgw_attrdispinfo *adip )
 	  if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 	    dsgw_emits( "\"" );
 	  }
-	  emit_value( adip->adi_vals[ i ], 
+	  emit_value( adip->adi_vals[ i ]->bv_val, 
 		    (( adip->adi_opts & DSGW_ATTROPT_NO_ENTITIES ) == 0 ));
 	  if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 	    dsgw_emits( "\"" );
@@ -1718,7 +1690,7 @@ str_display( struct dsgw_attrdispinfo *adip )
 	}
 
 	if ( adip->adi_htmltype != DSGW_ATTRHTML_HIDDEN &&
-		adip->adi_vals[ i + 1 ] != NULL ) {
+		adip->adi_vals[ i + 1 ] && adip->adi_vals[ i + 1 ]->bv_val ) {
 	    dsgw_emits( "<BR>\n" );
 	}
     }
@@ -1731,10 +1703,10 @@ ntuserid_display( struct dsgw_attrdispinfo *adip )
 {
     int		i;
 
-    for ( i = 0; adip->adi_vals[ i ] != NULL; ++i ) {
+    for ( i = 0; adip->adi_vals[ i ] && adip->adi_vals[ i ]->bv_val; ++i ) {
 	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv, 
 		  adip->adi_vals[ i ], adip->adi_vals[ i ] )) {
-            char *pch = adip->adi_vals[ i ];
+            char *pch = adip->adi_vals[ i ]->bv_val;
             if( pch ) {
 
 		if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
@@ -1748,7 +1720,7 @@ ntuserid_display( struct dsgw_attrdispinfo *adip )
 	    }
 	}
 
-	if ( adip->adi_vals[ i + 1 ] != NULL ) {
+	if ( adip->adi_vals[ i + 1 ] && adip->adi_vals[ i + 1 ]->bv_val ) {
 	    fputs( "<BR>\n", stdout );
 	}
     }
@@ -1785,8 +1757,8 @@ str_edit( struct dsgw_attrdispinfo *adip )
 	if ( adip->adi_vals == NULL ) {
 	    valcount = 0;
 	} else {
-	    for ( valcount = 0; adip->adi_vals[ valcount ] != NULL;
-		    ++valcount ) {
+	    for ( valcount = 0; adip->adi_vals[ valcount ] && 
+	                     adip->adi_vals[ valcount ]->bv_val; ++valcount ) {
 		;
 	    }
 	}
@@ -1797,7 +1769,9 @@ str_edit( struct dsgw_attrdispinfo *adip )
     case DSGW_ATTRHTML_TEXT:
     case DSGW_ATTRHTML_HIDDEN:
 	output_text_elements( adip->adi_argc, adip->adi_argv, adip->adi_attr,
-		adip->adi_vals, adip->adi_rdn, prefix, adip->adi_htmltype, textopts );
+	                      adip->adi_vals,
+			      adip->adi_rdn ? adip->adi_rdn->bv_val : NULL,
+	                      prefix, adip->adi_htmltype, textopts );
 	break;
 
     case DSGW_ATTRHTML_CHECKBOX:
@@ -1842,7 +1816,8 @@ mls_display( struct dsgw_attrdispinfo *adip )
 static void
 mls_edit( struct dsgw_attrdispinfo *adip )
 {
-    char		*prefix, **valscopy, *tval[ 2 ];
+    char		*prefix;
+    struct berval       **valscopy, *tval[ 2 ];
     int			i, valcount, adding, pre_idx, *lines;
     unsigned long	textopts;
 
@@ -1866,11 +1841,12 @@ mls_edit( struct dsgw_attrdispinfo *adip )
 	for ( valcount = 0; adip->adi_vals[ valcount ] != NULL; ++valcount ) {
 	    ;
 	}
-	valscopy = (char **)dsgw_ch_malloc( (valcount + 1) * sizeof( char * ));
+	valscopy = (struct berval **)dsgw_ch_malloc( (valcount + 1) * sizeof( char * ));
 	lines = (int *)dsgw_ch_malloc( valcount * sizeof( int ));
 	for ( i = 0; i < valcount; ++i ) {
-	    valscopy[ i ] = dsgw_mls_convertlines( adip->adi_vals[ i ], "\n",
-		    &lines[ i ], 0, 0 );
+	    valscopy[ i ]->bv_val = dsgw_mls_convertlines( adip->adi_vals[ i ],
+	                                             "\n", &lines[ i ], 0, 0 );
+	    valscopy[ i ]->bv_len = strlen(valscopy[ i ]->bv_val);
 	}
 	valscopy[ valcount ] = NULL;
     }
@@ -1900,7 +1876,7 @@ mls_edit( struct dsgw_attrdispinfo *adip )
     }
 
     if ( valscopy != NULL ) {
-	ldap_value_free( valscopy );
+	ldap_value_free_len( valscopy );
 	free( lines );
     }
 }
@@ -1915,12 +1891,16 @@ mls_edit( struct dsgw_attrdispinfo *adip )
  *    quote_html_specials flag) and NULL is returned.
  */
 char *
-dsgw_mls_convertlines( char *val, char *sep, int *linesp, int emitlines,
-	int quote_html_specials )
+dsgw_mls_convertlines( struct berval *val, char *sep, int *linesp, 
+                       int emitlines, int quote_html_specials )
 {
     char	*valcopy, *p, *q, *curline;
     int		i, c, lines, seplen;
 
+    if ( val == NULL ) {
+	return NULL;
+    }
+
     if ( sep == NULL ) {
 	sep = "";
 	seplen = 0;
@@ -1929,7 +1909,7 @@ dsgw_mls_convertlines( char *val, char *sep, int *linesp, int emitlines,
     }
 
     lines = 0;
-    for ( q = val; *q != '\0'; ++q ) {
+    for ( q = val->bv_val; *q != '\0'; ++q ) {
 	if ( *q == '$' ) {
 	    ++lines;
 	}
@@ -1939,7 +1919,7 @@ dsgw_mls_convertlines( char *val, char *sep, int *linesp, int emitlines,
 	*linesp = lines;
     }
 
-    valcopy = dsgw_ch_malloc( strlen( val ) + lines * seplen + 1 );
+    valcopy = dsgw_ch_malloc( strlen( val->bv_val ) + lines * seplen + 1 );
 
     /*
      * p points to the place we are copying to
@@ -1947,7 +1927,7 @@ dsgw_mls_convertlines( char *val, char *sep, int *linesp, int emitlines,
      * curline points to the start of the current line
      */
     p = curline = valcopy;
-    for ( q = val; *q != '\0'; ++q ) {
+    for ( q = val->bv_val; *q != '\0'; ++q ) {
 	if ( *q == '$' ) {		/* line separator */
 	    if ( emitlines ) {
 		*p = '\0';
@@ -2005,7 +1985,9 @@ static void
 dn_display( struct dsgw_attrdispinfo *adip )
 {
     int		i, j, len, dncomps;
-    char	*p, *staticlabel, *tmps = NULL, *label, *urlprefix, **rdns = NULL;
+    char	*p, *staticlabel, *tmps = NULL, *urlprefix, **rdns = NULL;
+    struct berval *label = NULL;
+    struct berval blabel;
 
     staticlabel = get_arg_by_name( DSGW_ATTRARG_LABEL, adip->adi_argc,
 	    adip->adi_argv );
@@ -2027,11 +2009,14 @@ dn_display( struct dsgw_attrdispinfo *adip )
     dsgw_log( "dn_display: urlprefix is %s\n", urlprefix );
 #endif
 
-    for ( i = 0; adip->adi_vals != NULL && adip->adi_vals[ i ] != NULL; ++i ) {
+    for ( i = 0; adip->adi_vals && adip->adi_vals[ i ] && 
+                 adip->adi_vals[ i ]->bv_val; ++i ) {
 	if ( staticlabel != NULL ) {
-	    label = staticlabel;
-	} else if ( !looks_like_dn( adip->adi_vals[ i ]) || 
-		( rdns = ldap_explode_dn( adip->adi_vals[ i ],
+	    blabel.bv_val = staticlabel;
+	    blabel.bv_len = strlen(staticlabel);
+	    label = &blabel;
+	} else if ( !looks_like_dn( adip->adi_vals[ i ]->bv_val ) || 
+		( rdns = dsgw_ldap_explode_dn( adip->adi_vals[ i ]->bv_val ,
 		( adip->adi_opts & DSGW_ATTROPT_DNTAGS ) == 0 )) == NULL ) {
 	    /* explode DN failed -- show entire DN */
 	    label = adip->adi_vals[ i ];
@@ -2042,7 +2027,9 @@ dn_display( struct dsgw_attrdispinfo *adip )
 		    ++ j ) {
 		len += ( 2 + strlen( rdns[ j ] ));	/* rdn + ", " */
 	    }
-	    label = p = tmps = dsgw_ch_malloc( len );
+	    blabel.bv_val = p = tmps = dsgw_ch_malloc( len );
+	    blabel.bv_len = len;
+	    label = &blabel;
 	    for ( j = 0; rdns[ j ] != NULL && ( dncomps == 0 || j < dncomps );
 		    ++ j ) {
 		if ( j > 0 ) {
@@ -2057,29 +2044,29 @@ dn_display( struct dsgw_attrdispinfo *adip )
 	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv, label,
 		adip->adi_vals[ i ] )) {
 	    if (( adip->adi_opts & DSGW_ATTROPT_NOLINK ) == 0 &&
-		    looks_like_dn( adip->adi_vals[ i ] )) {
+		    looks_like_dn( adip->adi_vals[ i ]->bv_val  )) {
 		if (( adip->adi_opts & DSGW_ATTROPT_DNPICKER ) != 0 ) {
 		    dsgw_emits( "<TR><TD>" );
 		}
 		/* Don't display a link for the rootdn */
-		if ( gc->gc_rootdn && dsgw_dn_cmp(adip->adi_vals[i], gc->gc_rootdn)) {
+		if ( gc->gc_rootdn && dsgw_dn_cmp(adip->adi_vals[i]->bv_val, gc->gc_rootdn)) {
 		  if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		    dsgw_emits( "\"" );
 		  }
-		  dsgw_emits( label );
+		  dsgw_emits( label->bv_val );
 		  if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		    dsgw_emits( "\"" );
 		  }
 		} else {
-		  dsgw_html_href( urlprefix, adip->adi_vals[ i ], label,
-				 adip->adi_vals[ i ],
+		  dsgw_html_href( urlprefix, adip->adi_vals[ i ]->bv_val, 
+				 label->bv_val, adip->adi_vals[ i ]->bv_val,
 				 get_arg_by_name( DSGW_ATTRARG_HREFEXTRA,
 						 adip->adi_argc, adip->adi_argv ));
 		}
 		if (( adip->adi_opts & DSGW_ATTROPT_DNPICKER ) != 0 ) {
 		    dsgw_emits( "</TD>\n<TD ALIGN=CENTER><INPUT TYPE=CHECKBOX " );
 		    dsgw_emitf( "VALUE=\"%s\" NAME=delete_%s ",
-			   adip->adi_vals[ i ], adip->adi_attr );
+			   adip->adi_vals[ i ]->bv_val, adip->adi_attr );
 		    dsgw_emitf( "onClick=\"aChg('%s');\"</TD>\n</TR>\n",
 			    adip->adi_attr );
 		}
@@ -2088,7 +2075,7 @@ dn_display( struct dsgw_attrdispinfo *adip )
 		dsgw_emits( "\"" );
 	      }
     
-	      emit_value( label,
+	      emit_value( label->bv_val,
 			(( adip->adi_opts & DSGW_ATTROPT_NO_ENTITIES ) == 0 ));
 	      if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		dsgw_emits( "\"" );
@@ -2106,7 +2093,7 @@ dn_display( struct dsgw_attrdispinfo *adip )
 	}
 
 	if ( rdns != NULL ) {
-	    ldap_value_free( rdns );
+	    dsgw_charray_free( rdns );
 	}
     }
 
@@ -2118,7 +2105,7 @@ dn_display( struct dsgw_attrdispinfo *adip )
 	dsgw_emitf( "var %s_values = new Object;\n", adip->adi_attr );
 	for ( i = 0; adip->adi_vals != NULL && adip->adi_vals[ i ] != NULL; ++i ) {
 	    char *edn;
-	    edn = dsgw_strdup_escaped( adip->adi_vals[ i ]);
+	    edn = dsgw_strdup_escaped( adip->adi_vals[ i ]->bv_val );
 	    dsgw_emitf( "%s_values[%d] = \"%s\";\n", adip->adi_attr, i,
 		    edn );
 	    free( edn );
@@ -2137,11 +2124,12 @@ mail_display( struct dsgw_attrdispinfo *adip )
 {
     int		i;
 
-    for ( i = 0; adip->adi_vals[ i ] != NULL; ++i ) {
+    for ( i = 0; adip->adi_vals[ i ] && adip->adi_vals[ i ]->bv_val; ++i ) {
 	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv,
 		    adip->adi_vals[ i ], adip->adi_vals[ i ] )) {
 	    if (( adip->adi_opts & DSGW_ATTROPT_NOLINK ) == 0 ) {
-		dsgw_html_href( "mailto:", adip->adi_vals[ i ], adip->adi_vals[ i ], NULL,
+		dsgw_html_href( "mailto:", adip->adi_vals[ i ]->bv_val,
+			adip->adi_vals[ i ]->bv_val, NULL,
 			get_arg_by_name( DSGW_ATTRARG_HREFEXTRA,
 				adip->adi_argc, adip->adi_argv ));
 	    } else {
@@ -2149,7 +2137,7 @@ mail_display( struct dsgw_attrdispinfo *adip )
 		dsgw_emits( "\"" );
 	      }
 	      
-	      emit_value( adip->adi_vals[ i ],
+	      emit_value( adip->adi_vals[ i ]->bv_val,
 			  (( adip->adi_opts & DSGW_ATTROPT_NO_ENTITIES ) == 0 ));
 	      if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		dsgw_emits( "\"" );
@@ -2171,20 +2159,23 @@ url_display( struct dsgw_attrdispinfo *adip )
 {
     int		i;
     char	*savep, *label;
+    struct berval blabel;
 
-    for ( i = 0; adip->adi_vals[ i ] != NULL; ++i ) {
-	if (( label = strchr( adip->adi_vals[ i ], ' ' )) == NULL ) {
-	    label = adip->adi_vals[ i ];
+    for ( i = 0; adip->adi_vals[ i ] && adip->adi_vals[ i ]->bv_val; ++i ) {
+	if (( label = strchr( adip->adi_vals[ i ]->bv_val, ' ' )) == NULL ) {
+	    label = adip->adi_vals[ i ]->bv_val;
 	    savep = NULL;
 	} else {
 	    savep = label;
 	    *label++ = '\0';
 	}
+	blabel.bv_val = label;
+	blabel.bv_len = strlen(label);
 
-	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv, label,
+	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv, &blabel,
 		adip->adi_vals[ i ] )) {
 	    if (( adip->adi_opts & DSGW_ATTROPT_NOLINK ) == 0 ) {
-		dsgw_html_href( NULL, adip->adi_vals[ i ], label, NULL,
+		dsgw_html_href( NULL, adip->adi_vals[ i ]->bv_val, label, NULL,
 			get_arg_by_name( DSGW_ATTRARG_HREFEXTRA,
 				adip->adi_argc, adip->adi_argv ));
 	    } else {
@@ -2192,7 +2183,7 @@ url_display( struct dsgw_attrdispinfo *adip )
 		    dsgw_emits( "\"" );
 		}
 		
-		emit_value( adip->adi_vals[ i ],
+		emit_value( adip->adi_vals[ i ]->bv_val,
 			(( adip->adi_opts & DSGW_ATTROPT_NO_ENTITIES ) == 0 ));
 		if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		    dsgw_emits( "\"" );
@@ -2219,6 +2210,7 @@ bool_display( struct dsgw_attrdispinfo *adip )
     int		boolval, pre_idx;
     char	*usestr, *truestr, *falsestr, *checked;
     char	*nameprefix, *onclick;
+    struct berval busestr;
 
     if ( adip->adi_vals == NULL || adip->adi_vals[ 0 ] == NULL ) {
       return;
@@ -2252,7 +2244,7 @@ bool_display( struct dsgw_attrdispinfo *adip )
 	falsestr = DSGW_ATTRARG_FALSESTR;
     }
 
-    boolval = ( toupper( adip->adi_vals[ 0 ][ 0 ] ) == 'T' );
+    boolval = ( toupper( *(adip->adi_vals[ 0 ]->bv_val) ) == 'T' );
 
     if ( adip->adi_htmltype == DSGW_ATTRHTML_RADIO ) {
 	dsgw_emitf( "<INPUT TYPE=\"radio\" NAME=\"%s%s\" "
@@ -2267,7 +2259,9 @@ bool_display( struct dsgw_attrdispinfo *adip )
 		boolval ? checked : "", onclick, truestr );
     } else {
 	usestr =  boolval ? truestr : falsestr;
-	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv, usestr,
+	busestr.bv_val = usestr;
+	busestr.bv_len = strlen(usestr);
+	if ( !did_output_as_special( adip->adi_argc, adip->adi_argv, &busestr,
 		adip->adi_vals[ 0 ] )) {
 	    if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		dsgw_emits( "\"" );
@@ -2305,7 +2299,7 @@ time_display( struct dsgw_attrdispinfo *adip )
 	    if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		dsgw_emits( "\"" );
 	    }
-	    dsgw_emits( time2text( adip->adi_vals[ i ],
+	    dsgw_emits( time2text( adip->adi_vals[ i ]->bv_val,
 		( adip->adi_opts & DSGW_ATTROPT_DATEONLY ) != 0 ) );
 	    if ((adip->adi_opts & DSGW_ATTROPT_QUOTED ) != 0 ) {
 		dsgw_emits( "\"" );
@@ -2325,13 +2319,17 @@ time_display( struct dsgw_attrdispinfo *adip )
  * return 0 if nothing was output or 1 if something was.
  */
 static int
-did_output_as_special( int argc, char **argv, char *label, char *val )
+did_output_as_special( int argc, char **argv, 
+                       struct berval *label, struct berval *val )
 {
     char	*href = NULL;
     char	*within = NULL;
     char	*script = NULL;
     char	*newval = NULL;
 
+    if (( NULL == val ) || ( NULL == label )) {
+	return( 0 );
+    }
     if (( href = get_arg_by_name( DSGW_ATTRARG_HREF, argc, argv )) == NULL &&
 	    ( within = get_arg_by_name( DSGW_ATTRARG_WITHIN, argc,
 	    argv )) == NULL &&
@@ -2341,12 +2339,12 @@ did_output_as_special( int argc, char **argv, char *label, char *val )
     }
 
     if ( within != NULL ) {
-	dsgw_substitute_and_output( within, "--value--", val, 1 );
+	dsgw_substitute_and_output( within, "--value--", val->bv_val, 1 );
     } else if (href != NULL) {
-	dsgw_html_href( NULL, href, label, val,
+	dsgw_html_href( NULL, href, label->bv_val, val->bv_val,
 		get_arg_by_name( DSGW_ATTRARG_HREFEXTRA, argc, argv ));
     } else if (script != NULL) {
-        newval = dsgw_strdup_escaped ( val );
+        newval = dsgw_strdup_escaped ( val->bv_val );
         if (newval != NULL && *newval != '\0') {
 		fputs( newval, stdout );
                 free( newval );
@@ -2576,11 +2574,11 @@ do_editbutton( char *dn, char *encodeddn, int argc, char **argv )
 	buttonlabel = XP_GetClientStr(DBT_edit_);
     }
 
-    if (( rdns = ldap_explode_dn( dn, 1 )) != NULL ) {
+    if (( rdns = dsgw_ldap_explode_dn( dn, 1 )) != NULL ) {
 	dsgw_emitf(
 		"<INPUT TYPE=\"hidden\" NAME=\"authhint\" VALUE=\"%s\">\n",
 		rdns[ 0 ] );
-	ldap_value_free( rdns );
+	dsgw_charray_free( rdns );
     }
 
     dsgw_emitf( "<INPUT TYPE=\"hidden\" NAME=\"authdesturl\">\n"
@@ -2910,22 +2908,26 @@ condition_is_true( int argc, char **argv, void *arg )
       return((gc->gc_aimpresence == 1) && ((tip->dsti_options & DSGW_DISPLAY_OPT_ADDING ) == 0));
     }
 
+    if ( strcasecmp( argv[0], DSGW_COND_HAVEAIMSCHEMA ) == 0 ) {
+      return((gc->gc_haveaimschema == 1) && ((tip->dsti_options & DSGW_DISPLAY_OPT_ADDING ) == 0));
+    }
+
     if ( strcasecmp( argv[0], DSGW_COND_ATTRHASVALUES ) == 0 ) {
 	/*
 	 * format of IF statment is:
 	 *    <-- IF "AttributeHasValues" "ATTRIBUTE" "MINIMUM_COUNT" -->
 	 * MINIMUM_COUNT is an optional number.
 	 */
-	char	**vals;
+	struct berval	**vals;
 	int	rc, minimum;
 
 	if ( argc < 2 || tip->dsti_entry == NULL ||
-		( vals = (char **) ldap_get_values( tip->dsti_ld, tip->dsti_entry,
-		argv[1])) == NULL ) {
+		( vals = ldap_get_values_len( tip->dsti_ld,
+		                       tip->dsti_entry, argv[1])) == NULL ) {
 	    /* check "attrsonly" information if applicable */
 	    if ( argc < 3 && tip->dsti_attrsonly_entry != NULL ) {
-		(void)ldap_get_values( tip->dsti_ld, tip->dsti_attrsonly_entry,	argv[1]);
-		if ( ldap_get_lderrno( tip->dsti_ld, NULL, NULL )
+		(void)ldap_get_values_len( tip->dsti_ld, tip->dsti_attrsonly_entry,	argv[1]);
+		if ( dsgw_ldap_get_lderrno( tip->dsti_ld, NULL, NULL )
 			== LDAP_SUCCESS ) {
 		    return( 1 );
 		}
@@ -2933,8 +2935,8 @@ condition_is_true( int argc, char **argv, void *arg )
 	    return( 0 );
 	}
 	minimum = ( argc < 3 ) ? 1 : atoi( argv[ 2 ] );
-	rc = ( minimum <= 1 || ldap_count_values( vals ) >= minimum );
-	ldap_value_free( vals );
+	rc = ( minimum <= 1 || ldap_count_values_len( vals ) >= minimum );
+	ldap_value_free_len( vals );
 	return( rc );
     }
 
@@ -2943,12 +2945,14 @@ condition_is_true( int argc, char **argv, void *arg )
 	 * format of IF statment is:
 	 *    <-- IF "AttributeHasThisValue" "ATTRIBUTE" "SYNTAX" "VALUE" -->
 	 */
-	char			**vals;
+	struct berval	**vals;
 	int			i, rc;
 	struct attr_handler	*ahp;
+	struct berval 		thisarg;
+	struct berval 		*thisargp;
 
 	if ( argc < 4 ||  tip->dsti_entry == NULL ||
-	     ( vals = (char **) ldap_get_values( tip->dsti_ld, tip->dsti_entry,
+	     ( vals = ldap_get_values_len( tip->dsti_ld, tip->dsti_entry,
 				       argv[1])) == NULL ) {
 	    return( 0 );
 	}
@@ -2958,14 +2962,17 @@ condition_is_true( int argc, char **argv, void *arg )
 	}
 
 	rc = 0;
-	for ( i = 0; vals[ i ] != NULL; ++i ) {
-	    if ( dsgw_valcmp(ahp->ath_compare)( (const char **)&vals[i],
-		    (const char **)&argv[3] ) == 0 ) {
+	thisarg.bv_val = argv[3]?argv[3]:"";
+	thisarg.bv_len = argv[3]?strlen(argv[3]):0;
+	thisargp = &thisarg;
+	for ( i = 0; vals[ i ] && vals[ i ]->bv_val; ++i ) {
+	    if ( dsgw_valcmp(ahp->ath_compare)( (const struct berval **)&vals[i],
+		    (const struct berval **)&thisargp ) == 0 ) {
 		rc = 1;
 		break;
 	    }
 	}
-	ldap_value_free( vals );
+	ldap_value_free_len( vals );
 	return( rc );
     }
 
@@ -2983,22 +2990,20 @@ condition_is_true( int argc, char **argv, void *arg )
  *              possible language or phonetic tags.
  *              pass in something like "cn" and this function
  *              will return all cn's, tagged or not.
- *              If binary_value is 1, then it'll handle
- *              everything as binary values.
  *
  * Author: RJP
  *
  */
-static char ** 
+static struct berval ** 
 dsgw_get_values( LDAP *ld, LDAPMessage *entry, 
-		 const char *target, int binary_value ) 
+		 const char *target ) 
 {
     BerElement  *ber        = NULL;
     char        *attr       = NULL;
     char        *new_target = NULL;
     int          new_target_size = 0;
-    char       **val_youse  = NULL;
-    char       **temp_vals  = NULL;
+    struct berval **val_youse  = NULL;
+    struct berval **temp_vals  = NULL;
     int          i          = 0;
     int          j          = 0;
     int          temp_val_count = 0;
@@ -3019,11 +3024,7 @@ dsgw_get_values( LDAP *ld, LDAPMessage *entry,
 	/* If the "target;" matches the attribute name, get the values*/
 	if ( strcasecmp(attr, target) == 0 ||
 	     strncasecmp (attr, new_target, new_target_size) == 0) {
-	    if (binary_value) {
-		temp_vals = (char **) ldap_get_values_len( ld, entry, attr );
-	    } else {
-		temp_vals = (char **) ldap_get_values( ld, entry, attr );
-	    }
+	    temp_vals =  ldap_get_values_len( ld, entry, attr );
 	    
 	    if (temp_vals == NULL) {
 		continue;
@@ -3031,7 +3032,7 @@ dsgw_get_values( LDAP *ld, LDAPMessage *entry,
 	    
 	    /* Find the next open spot in val_youse*/
 	    if (val_youse) {
-		for (; val_youse[i] != NULL; i++) ;
+		for (; val_youse[i] && val_youse[i]->bv_val; i++) ;
 	    }
 
 	    /* Count the number of values in temp_vals */
@@ -3039,7 +3040,7 @@ dsgw_get_values( LDAP *ld, LDAPMessage *entry,
 		 temp_val_count++);
 	    
 	    /* Realloc */
-	    val_youse = (char **) dsgw_ch_realloc (val_youse, sizeof(char *) * (temp_val_count + i + 1) );
+	    val_youse = (struct berval **) dsgw_ch_realloc (val_youse, sizeof(struct berval *) * (temp_val_count + i + 1) );
 	    
 	    /* Start there and copy over the pointers from temp_vals */
 	    for (j = 0; j < temp_val_count; j++, i++) {
@@ -3057,7 +3058,7 @@ dsgw_get_values( LDAP *ld, LDAPMessage *entry,
     
     if ( ber != NULL ) {
 	
-	ldap_ber_free( ber, 0 );
+	ber_free( ber, 0 );
 	
     }
     
@@ -3066,40 +3067,6 @@ dsgw_get_values( LDAP *ld, LDAPMessage *entry,
     return(val_youse);
 }
 
-/*
- * Function: dsgw_value_free
- *
- * Returns: nothing
- *
- * Description: frees a half libldap and half dsge malloc'd array.
- *              Sorry. This really sucks, I know, but I didn't
- *              want to copy all that data around.
- *
- * Author: RJP
- *
- */
-static void
-dsgw_value_free( void **ldvals, int binary ) 
-{
-    int i;
-
-    for (i = 0; ldvals[i] != NULL; i ++) {
-	if (binary) {
-	    struct berval *delete_me = NULL;
-
-	    delete_me = (struct berval *) ldvals[i];
-	    
-	    ldap_memfree(delete_me->bv_val);
-	    ldap_memfree(delete_me);
-	} else {
-	    ldap_memfree (ldvals[i]);
-	}
-    }
-    
-    free(ldvals);
-	
-    
-}
 /*
  * Function: dsgw_time
  *
